@@ -17,9 +17,9 @@
 This module provides two helper methods `dumps` and `loads` that wrap the
 native :mod:`json` methods and provide explicit BSON conversion to and from
 JSON. :class:`~bson.json_util.JSONOptions` provides a way to control how JSON
-is emitted and parsed, with the default being the legacy PyMongo format.
-:mod:`~bson.json_util` can also generate Canonical or Relaxed `Extended JSON`_
-when :const:`CANONICAL_JSON_OPTIONS` or :const:`RELAXED_JSON_OPTIONS` is
+is emitted and parsed, with the default being the Relaxed Extended JSON format.
+:mod:`~bson.json_util` can also generate Canonical or legacy `Extended JSON`_
+when :const:`CANONICAL_JSON_OPTIONS` or :const:`LEGACY_JSON_OPTIONS` is
 provided, respectively.
 
 .. _Extended JSON: https://github.com/mongodb/specifications/blob/master/source/extended-json.rst
@@ -30,9 +30,9 @@ Example usage (deserialization):
 
    >>> from bson.json_util import loads
    >>> loads('[{"foo": [1, 2]}, {"bar": {"hello": "world"}}, {"code": {"$scope": {}, "$code": "function x() { return 1; }"}}, {"bin": {"$type": "80", "$binary": "AQIDBA=="}}]')
-   [{u'foo': [1, 2]}, {u'bar': {u'hello': u'world'}}, {u'code': Code('function x() { return 1; }', {})}, {u'bin': Binary('...', 128)}]
+   [{'foo': [1, 2]}, {'bar': {'hello': 'world'}}, {'code': Code('function x() { return 1; }', {})}, {'bin': Binary(b'...', 128)}]
 
-Example usage (serialization):
+Example usage with :const:`RELAXED_JSON_OPTIONS` (the default):
 
 .. doctest::
 
@@ -40,9 +40,9 @@ Example usage (serialization):
    >>> from bson.json_util import dumps
    >>> dumps([{'foo': [1, 2]},
    ...        {'bar': {'hello': 'world'}},
-   ...        {'code': Code("function x() { return 1; }", {})},
+   ...        {'code': Code("function x() { return 1; }")},
    ...        {'bin': Binary(b"\x01\x02\x03\x04")}])
-   '[{"foo": [1, 2]}, {"bar": {"hello": "world"}}, {"code": {"$code": "function x() { return 1; }", "$scope": {}}}, {"bin": {"$binary": "AQIDBA==", "$type": "00"}}]'
+   '[{"foo": [1, 2]}, {"bar": {"hello": "world"}}, {"code": {"$code": "function x() { return 1; }"}}, {"bin": {"$binary": {"base64": "AQIDBA==", "subType": "00"}}}]'
 
 Example usage (with :const:`CANONICAL_JSON_OPTIONS`):
 
@@ -57,18 +57,18 @@ Example usage (with :const:`CANONICAL_JSON_OPTIONS`):
    ...       json_options=CANONICAL_JSON_OPTIONS)
    '[{"foo": [{"$numberInt": "1"}, {"$numberInt": "2"}]}, {"bar": {"hello": "world"}}, {"code": {"$code": "function x() { return 1; }"}}, {"bin": {"$binary": {"base64": "AQIDBA==", "subType": "00"}}}]'
 
-Example usage (with :const:`RELAXED_JSON_OPTIONS`):
+Example usage (with :const:`LEGACY_JSON_OPTIONS`):
 
 .. doctest::
 
    >>> from bson import Binary, Code
-   >>> from bson.json_util import dumps, RELAXED_JSON_OPTIONS
+   >>> from bson.json_util import dumps, LEGACY_JSON_OPTIONS
    >>> dumps([{'foo': [1, 2]},
    ...        {'bar': {'hello': 'world'}},
-   ...        {'code': Code("function x() { return 1; }")},
+   ...        {'code': Code("function x() { return 1; }", {})},
    ...        {'bin': Binary(b"\x01\x02\x03\x04")}],
-   ...       json_options=RELAXED_JSON_OPTIONS)
-   '[{"foo": [1, 2]}, {"bar": {"hello": "world"}}, {"code": {"$code": "function x() { return 1; }"}}, {"bin": {"$binary": {"base64": "AQIDBA==", "subType": "00"}}}]'
+   ...       json_options=LEGACY_JSON_OPTIONS)
+   '[{"foo": [1, 2]}, {"bar": {"hello": "world"}}, {"code": {"$code": "function x() { return 1; }", "$scope": {}}}, {"bin": {"$binary": "AQIDBA==", "$type": "00"}}]'
 
 Alternatively, you can manually pass the `default` to :func:`json.dumps`.
 It won't handle :class:`~bson.binary.Binary` and :class:`~bson.code.Code`
@@ -84,25 +84,6 @@ but it will be faster as there is less recursion.
    Extended JSON converter for Python built on top of
    `libbson <https://github.com/mongodb/libbson>`_. `python-bsonjs` works best
    with PyMongo when using :class:`~bson.raw_bson.RawBSONDocument`.
-
-.. versionchanged:: 2.8
-   The output format for :class:`~bson.timestamp.Timestamp` has changed from
-   '{"t": <int>, "i": <int>}' to '{"$timestamp": {"t": <int>, "i": <int>}}'.
-   This new format will be decoded to an instance of
-   :class:`~bson.timestamp.Timestamp`. The old format will continue to be
-   decoded to a python dict as before. Encoding to the old format is no longer
-   supported as it was never correct and loses type information.
-   Added support for $numberLong and $undefined - new in MongoDB 2.6 - and
-   parsing $date in ISO-8601 format.
-
-.. versionchanged:: 2.7
-   Preserves order when rendering SON, Timestamp, Code, Binary, and DBRef
-   instances.
-
-.. versionchanged:: 2.3
-   Added dumps and loads helpers to automatically handle conversion to and
-   from json and supports :class:`~bson.binary.Binary` and
-   :class:`~bson.code.Code`
 """
 
 import base64
@@ -112,8 +93,6 @@ import math
 import re
 import uuid
 
-from pymongo.errors import ConfigurationError
-
 import bson
 from bson import EPOCH_AWARE, RE_TYPE, SON
 from bson.binary import (Binary, UuidRepresentation, ALL_UUID_SUBTYPES,
@@ -126,8 +105,6 @@ from bson.int64 import Int64
 from bson.max_key import MaxKey
 from bson.min_key import MinKey
 from bson.objectid import ObjectId
-from bson.py3compat import (PY3, iteritems, integer_types, string_type,
-                            text_type)
 from bson.regex import Regex
 from bson.timestamp import Timestamp
 from bson.tz_util import utc
@@ -142,9 +119,6 @@ _RE_OPT_TABLE = {
     "x": re.X,
 }
 
-# Dollar-prefixed keys which may appear in DBRefs.
-_DBREF_KEYS = frozenset(['$id', '$ref', '$db'])
-
 
 class DatetimeRepresentation:
     LEGACY = 0
@@ -246,11 +220,11 @@ class JSONOptions(CodecOptions):
         :class:`collections.MutableMapping`. Defaults to :class:`dict`.
       - `uuid_representation`: The :class:`~bson.binary.UuidRepresentation`
         to use when encoding and decoding instances of :class:`uuid.UUID`.
-        Defaults to :const:`~bson.binary.UuidRepresentation.PYTHON_LEGACY`.
+        Defaults to :const:`~bson.binary.UuidRepresentation.UNSPECIFIED`.
       - `tz_aware`: If ``True``, MongoDB Extended JSON's *Strict mode* type
         `Date` will be decoded to timezone aware instances of
         :class:`datetime.datetime`. Otherwise they will be naive. Defaults
-        to ``True``.
+        to ``False``.
       - `tzinfo`: A :class:`datetime.tzinfo` subclass that specifies the
         timezone from which :class:`~datetime.datetime` objects should be
         decoded. Defaults to :const:`~bson.tz_util.utc`.
@@ -259,46 +233,84 @@ class JSONOptions(CodecOptions):
 
     .. seealso:: The specification for Relaxed and Canonical `Extended JSON`_.
 
-    .. versionadded:: 3.4
+    .. versionchanged:: 4.0
+       The default for `json_mode` was changed from :const:`JSONMode.LEGACY`
+       to :const:`JSONMode.RELAXED`.
+       The default for `uuid_representation` was changed from
+       :const:`~bson.binary.UuidRepresentation.PYTHON_LEGACY` to
+       :const:`~bson.binary.UuidRepresentation.UNSPECIFIED`.
 
     .. versionchanged:: 3.5
        Accepts the optional parameter `json_mode`.
 
+    .. versionchanged:: 4.0
+       Changed default value of `tz_aware` to False.
     """
 
-    def __new__(cls, strict_number_long=False,
-                datetime_representation=DatetimeRepresentation.LEGACY,
-                strict_uuid=False, json_mode=JSONMode.LEGACY,
+    def __new__(cls, strict_number_long=None,
+                datetime_representation=None,
+                strict_uuid=None, json_mode=JSONMode.RELAXED,
                 *args, **kwargs):
-        kwargs["tz_aware"] = kwargs.get("tz_aware", True)
+        kwargs["tz_aware"] = kwargs.get("tz_aware", False)
         if kwargs["tz_aware"]:
             kwargs["tzinfo"] = kwargs.get("tzinfo", utc)
         if datetime_representation not in (DatetimeRepresentation.LEGACY,
                                            DatetimeRepresentation.NUMBERLONG,
-                                           DatetimeRepresentation.ISO8601):
-            raise ConfigurationError(
+                                           DatetimeRepresentation.ISO8601,
+                                           None):
+            raise ValueError(
                 "JSONOptions.datetime_representation must be one of LEGACY, "
                 "NUMBERLONG, or ISO8601 from DatetimeRepresentation.")
         self = super(JSONOptions, cls).__new__(cls, *args, **kwargs)
         if json_mode not in (JSONMode.LEGACY,
                              JSONMode.RELAXED,
                              JSONMode.CANONICAL):
-            raise ConfigurationError(
+            raise ValueError(
                 "JSONOptions.json_mode must be one of LEGACY, RELAXED, "
                 "or CANONICAL from JSONMode.")
         self.json_mode = json_mode
         if self.json_mode == JSONMode.RELAXED:
+            if strict_number_long:
+                raise ValueError(
+                    "Cannot specify strict_number_long=True with"
+                    " JSONMode.RELAXED")
+            if datetime_representation not in (None,
+                                               DatetimeRepresentation.ISO8601):
+                raise ValueError(
+                    "datetime_representation must be DatetimeRepresentation."
+                    "ISO8601 or omitted with JSONMode.RELAXED")
+            if strict_uuid not in (None, True):
+                raise ValueError(
+                    "Cannot specify strict_uuid=False with JSONMode.RELAXED")
             self.strict_number_long = False
             self.datetime_representation = DatetimeRepresentation.ISO8601
             self.strict_uuid = True
         elif self.json_mode == JSONMode.CANONICAL:
+            if strict_number_long not in (None, True):
+                raise ValueError(
+                    "Cannot specify strict_number_long=False with"
+                    " JSONMode.RELAXED")
+            if datetime_representation not in (
+                    None, DatetimeRepresentation.NUMBERLONG):
+                raise ValueError(
+                    "datetime_representation must be DatetimeRepresentation."
+                    "NUMBERLONG or omitted with JSONMode.RELAXED")
+            if strict_uuid not in (None, True):
+                raise ValueError(
+                    "Cannot specify strict_uuid=False with JSONMode.RELAXED")
             self.strict_number_long = True
             self.datetime_representation = DatetimeRepresentation.NUMBERLONG
             self.strict_uuid = True
-        else:
-            self.strict_number_long = strict_number_long
-            self.datetime_representation = datetime_representation
-            self.strict_uuid = strict_uuid
+        else:  # JSONMode.LEGACY
+            self.strict_number_long = False
+            self.datetime_representation = DatetimeRepresentation.LEGACY
+            self.strict_uuid = False
+            if strict_number_long is not None:
+                self.strict_number_long = strict_number_long
+            if datetime_representation is not None:
+                self.datetime_representation = datetime_representation
+            if strict_uuid is not None:
+                self.strict_uuid = strict_uuid
         return self
 
     def _arguments_repr(self):
@@ -311,6 +323,36 @@ class JSONOptions(CodecOptions):
                     self.json_mode,
                     super(JSONOptions, self)._arguments_repr()))
 
+    def _options_dict(self):
+        # TODO: PYTHON-2442 use _asdict() instead
+        options_dict = super(JSONOptions, self)._options_dict()
+        options_dict.update({
+            'strict_number_long': self.strict_number_long,
+            'datetime_representation': self.datetime_representation,
+            'strict_uuid': self.strict_uuid,
+            'json_mode': self.json_mode})
+        return options_dict
+
+    def with_options(self, **kwargs):
+        """
+        Make a copy of this JSONOptions, overriding some options::
+
+            >>> from bson.json_util import CANONICAL_JSON_OPTIONS
+            >>> CANONICAL_JSON_OPTIONS.tz_aware
+            True
+            >>> json_options = CANONICAL_JSON_OPTIONS.with_options(tz_aware=False, tzinfo=None)
+            >>> json_options.tz_aware
+            False
+
+        .. versionadded:: 3.12
+        """
+        opts = self._options_dict()
+        for opt in ('strict_number_long', 'datetime_representation',
+                    'strict_uuid', 'json_mode'):
+            opts[opt] = kwargs.get(opt, getattr(self, opt))
+        opts.update(kwargs)
+        return JSONOptions(**opts)
+
 
 LEGACY_JSON_OPTIONS = JSONOptions(json_mode=JSONMode.LEGACY)
 """:class:`JSONOptions` for encoding to PyMongo's legacy JSON format.
@@ -320,15 +362,6 @@ LEGACY_JSON_OPTIONS = JSONOptions(json_mode=JSONMode.LEGACY)
 .. versionadded:: 3.5
 """
 
-DEFAULT_JSON_OPTIONS = LEGACY_JSON_OPTIONS
-"""The default :class:`JSONOptions` for JSON encoding/decoding.
-
-The same as :const:`LEGACY_JSON_OPTIONS`. This will change to
-:const:`RELAXED_JSON_OPTIONS` in a future release.
-
-.. versionadded:: 3.4
-"""
-
 CANONICAL_JSON_OPTIONS = JSONOptions(json_mode=JSONMode.CANONICAL)
 """:class:`JSONOptions` for Canonical Extended JSON.
 
@@ -345,18 +378,16 @@ RELAXED_JSON_OPTIONS = JSONOptions(json_mode=JSONMode.RELAXED)
 .. versionadded:: 3.5
 """
 
-STRICT_JSON_OPTIONS = JSONOptions(
-    strict_number_long=True,
-    datetime_representation=DatetimeRepresentation.ISO8601,
-    strict_uuid=True)
-"""**DEPRECATED** - :class:`JSONOptions` for MongoDB Extended JSON's *Strict
-mode* encoding.
+DEFAULT_JSON_OPTIONS = RELAXED_JSON_OPTIONS
+"""The default :class:`JSONOptions` for JSON encoding/decoding.
+
+The same as :const:`RELAXED_JSON_OPTIONS`.
+
+.. versionchanged:: 4.0
+   Changed from :const:`LEGACY_JSON_OPTIONS` to
+   :const:`RELAXED_JSON_OPTIONS`.
 
 .. versionadded:: 3.4
-
-.. versionchanged:: 3.5
-   Deprecated. Use :const:`RELAXED_JSON_OPTIONS` or
-   :const:`CANONICAL_JSON_OPTIONS` instead.
 """
 
 
@@ -371,12 +402,12 @@ def dumps(obj, *args, **kwargs):
         encoding of MongoDB Extended JSON types. Defaults to
         :const:`DEFAULT_JSON_OPTIONS`.
 
+    .. versionchanged:: 4.0
+       Now outputs MongoDB Relaxed Extended JSON by default (using
+       :const:`DEFAULT_JSON_OPTIONS`).
+
     .. versionchanged:: 3.4
        Accepts optional parameter `json_options`. See :class:`JSONOptions`.
-
-    .. versionchanged:: 2.7
-       Preserves order when rendering SON, Timestamp, Code, Binary, and DBRef
-       instances.
     """
     json_options = kwargs.pop("json_options", DEFAULT_JSON_OPTIONS)
     return json.dumps(_json_convert(obj, json_options), *args, **kwargs)
@@ -413,10 +444,10 @@ def _json_convert(obj, json_options=DEFAULT_JSON_OPTIONS):
     """Recursive helper method that converts BSON types so they can be
     converted into json.
     """
-    if hasattr(obj, 'iteritems') or hasattr(obj, 'items'):  # PY3 support
+    if hasattr(obj, 'items'):
         return SON(((k, _json_convert(v, json_options))
-                    for k, v in iteritems(obj)))
-    elif hasattr(obj, '__iter__') and not isinstance(obj, (text_type, bytes)):
+                    for k, v in obj.items()))
+    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes)):
         return list((_json_convert(v, json_options) for v in obj))
     try:
         return default(obj, json_options)
@@ -431,7 +462,9 @@ def object_pairs_hook(pairs, json_options=DEFAULT_JSON_OPTIONS):
 def object_hook(dct, json_options=DEFAULT_JSON_OPTIONS):
     if "$oid" in dct:
         return _parse_canonical_oid(dct)
-    if "$ref" in dct:
+    if (isinstance(dct.get('$ref'), str) and
+            "$id" in dct and
+            isinstance(dct.get('$db'), (str, type(None)))):
         return _parse_canonical_dbref(dct)
     if "$date" in dct:
         return _parse_canonical_datetime(dct, json_options)
@@ -488,6 +521,8 @@ def _parse_legacy_uuid(doc, json_options):
     """Decode a JSON legacy $uuid to Python UUID."""
     if len(doc) != 1:
         raise TypeError('Bad $uuid, extra field(s): %s' % (doc,))
+    if not isinstance(doc["$uuid"], str):
+        raise TypeError('$uuid must be a string: %s' % (doc,))
     if json_options.uuid_representation == UuidRepresentation.UNSPECIFIED:
         return Binary.from_uuid(uuid.UUID(doc["$uuid"]))
     else:
@@ -510,7 +545,7 @@ def _binary_or_uuid(data, subtype, json_options):
             uuid_representation = UuidRepresentation.PYTHON_LEGACY
         return binary_value.as_uuid(uuid_representation)
 
-    if PY3 and subtype == 0:
+    if subtype == 0:
         return data
     return Binary(data, subtype)
 
@@ -529,9 +564,9 @@ def _parse_canonical_binary(doc, json_options):
     binary = doc["$binary"]
     b64 = binary["base64"]
     subtype = binary["subType"]
-    if not isinstance(b64, string_type):
+    if not isinstance(b64, str):
         raise TypeError('$binary base64 must be a string: %s' % (doc,))
-    if not isinstance(subtype, string_type) or len(subtype) > 2:
+    if not isinstance(subtype, str) or len(subtype) > 2:
         raise TypeError('$binary subType must be a string at most 2 '
                         'characters: %s' % (doc,))
     if len(binary) != 2:
@@ -548,7 +583,7 @@ def _parse_canonical_datetime(doc, json_options):
     if len(doc) != 1:
         raise TypeError('Bad $date, extra field(s): %s' % (doc,))
     # mongoexport 2.6 and newer
-    if isinstance(dtm, string_type):
+    if isinstance(dtm, str):
         # Parse offset
         if dtm[-1] == 'Z':
             dt = dtm[:-1]
@@ -613,7 +648,7 @@ def _parse_canonical_symbol(doc):
     symbol = doc['$symbol']
     if len(doc) != 1:
         raise TypeError('Bad $symbol, extra field(s): %s' % (doc,))
-    return text_type(symbol)
+    return str(symbol)
 
 
 def _parse_canonical_code(doc):
@@ -632,15 +667,15 @@ def _parse_canonical_regex(doc):
     if len(regex) != 2:
         raise TypeError('Bad $regularExpression must include only "pattern"'
                         'and "options" components: %s' % (doc,))
-    return Regex(regex['pattern'], regex['options'])
+    opts = regex['options']
+    if not isinstance(opts, str):
+        raise TypeError('Bad $regularExpression options, options must be '
+                        'string, was type %s' % (type(opts)))
+    return Regex(regex['pattern'], opts)
 
 
 def _parse_canonical_dbref(doc):
     """Decode a JSON DBRef to bson.dbref.DBRef."""
-    for key in doc:
-        if key.startswith('$') and key not in _DBREF_KEYS:
-            # Other keys start with $, so dct cannot be parsed as a DBRef.
-            return doc
     return DBRef(doc.pop('$ref'), doc.pop('$id'),
                  database=doc.pop('$db', None), **doc)
 
@@ -672,7 +707,7 @@ def _parse_canonical_int32(doc):
     i_str = doc['$numberInt']
     if len(doc) != 1:
         raise TypeError('Bad $numberInt, extra field(s): %s' % (doc,))
-    if not isinstance(i_str, string_type):
+    if not isinstance(i_str, str):
         raise TypeError('$numberInt must be string: %s' % (doc,))
     return int(i_str)
 
@@ -690,7 +725,7 @@ def _parse_canonical_double(doc):
     d_str = doc['$numberDouble']
     if len(doc) != 1:
         raise TypeError('Bad $numberDouble, extra field(s): %s' % (doc,))
-    if not isinstance(d_str, string_type):
+    if not isinstance(d_str, str):
         raise TypeError('$numberDouble must be string: %s' % (doc,))
     return float(d_str)
 
@@ -700,7 +735,7 @@ def _parse_canonical_decimal128(doc):
     d_str = doc['$numberDecimal']
     if len(doc) != 1:
         raise TypeError('Bad $numberDecimal, extra field(s): %s' % (doc,))
-    if not isinstance(d_str, string_type):
+    if not isinstance(d_str, str):
         raise TypeError('$numberDecimal must be string: %s' % (doc,))
     return Decimal128(d_str)
 
@@ -777,7 +812,7 @@ def default(obj, json_options=DEFAULT_JSON_OPTIONS):
             flags += "u"
         if obj.flags & re.VERBOSE:
             flags += "x"
-        if isinstance(obj.pattern, text_type):
+        if isinstance(obj.pattern, str):
             pattern = obj.pattern
         else:
             pattern = obj.pattern.decode('utf-8')
@@ -799,7 +834,7 @@ def default(obj, json_options=DEFAULT_JSON_OPTIONS):
             ('$scope', _json_convert(obj.scope, json_options))])
     if isinstance(obj, Binary):
         return _encode_binary(obj, obj.subtype, json_options)
-    if PY3 and isinstance(obj, bytes):
+    if isinstance(obj, bytes):
         return _encode_binary(obj, 0, json_options)
     if isinstance(obj, uuid.UUID):
         if json_options.strict_uuid:
@@ -813,10 +848,10 @@ def default(obj, json_options=DEFAULT_JSON_OPTIONS):
     if isinstance(obj, bool):
         return obj
     if (json_options.json_mode == JSONMode.CANONICAL and
-            isinstance(obj, integer_types)):
+            isinstance(obj, int)):
         if -2 ** 31 <= obj < 2 ** 31:
-            return {'$numberInt': text_type(obj)}
-        return {'$numberLong': text_type(obj)}
+            return {'$numberInt': str(obj)}
+        return {'$numberLong': str(obj)}
     if json_options.json_mode != JSONMode.LEGACY and isinstance(obj, float):
         if math.isnan(obj):
             return {'$numberDouble': 'NaN'}
@@ -825,7 +860,6 @@ def default(obj, json_options=DEFAULT_JSON_OPTIONS):
             return {'$numberDouble': representation}
         elif json_options.json_mode == JSONMode.CANONICAL:
             # repr() will return the shortest string guaranteed to produce the
-            # original value, when float() is called on it. str produces a
-            # shorter string in Python 2.
-            return {'$numberDouble': text_type(repr(obj))}
+            # original value, when float() is called on it.
+            return {'$numberDouble': str(repr(obj))}
     raise TypeError("%r is not JSON serializable" % obj)
