@@ -14,23 +14,26 @@
 
 """Tools for representing files stored in GridFS."""
 import datetime
-import hashlib
+import io
 import math
 import os
 
+from bson.int64 import Int64
 from bson.son import SON
 from bson.binary import Binary
 from bson.objectid import ObjectId
-from bson.py3compat import text_type, StringIO
-from gridfs.errors import CorruptGridFile, FileExists, NoFile
 from pymongo import ASCENDING
 from pymongo.collection import Collection
 from pymongo.cursor import Cursor
 from pymongo.errors import (ConfigurationError,
+                            CursorNotFound,
                             DuplicateKeyError,
+                            InvalidOperation,
                             OperationFailure)
 from pymongo.read_preferences import ReadPreference
 
+from gridfs.errors import CorruptGridFile, FileExists, NoFile
+
 try:
     _SEEK_SET = os.SEEK_SET
     _SEEK_CUR = os.SEEK_CUR
@@ -96,11 +99,22 @@ def _grid_out_property(field_name, docstring):
     return property(getter, doc=docstring)
 
 
+def _clear_entity_type_registry(entity, **kwargs):
+    """Clear the given database/collection object's type registry."""
+    codecopts = entity.codec_options.with_options(type_registry=None)
+    return entity.with_options(codec_options=codecopts, **kwargs)
+
+
+def _disallow_transactions(session):
+    if session and session.in_transaction:
+        raise InvalidOperation(
+            'GridFS does not support multi-document transactions')
+
+
 class GridIn(object):
     """Class to write data to GridFS.
     """
-    def __init__(
-            self, root_collection, session=None, disable_md5=False, **kwargs):
+    def __init__(self, root_collection, session=None, **kwargs):
         """Write a file to GridFS
 
         Application developers should generally not need to
@@ -128,23 +142,23 @@ class GridIn(object):
           - ``"chunkSize"`` or ``"chunk_size"``: size of each of the
             chunks, in bytes (default: 255 kb)
 
-          - ``"encoding"``: encoding used for this file. In Python 2,
-            any :class:`unicode` that is written to the file will be
-            converted to a :class:`str`. In Python 3, any :class:`str`
-            that is written to the file will be converted to
-            :class:`bytes`.
+          - ``"encoding"``: encoding used for this file. Any :class:`str`
+            that is written to the file will be converted to :class:`bytes`.
 
         :Parameters:
           - `root_collection`: root collection to write to
           - `session` (optional): a
             :class:`~pymongo.client_session.ClientSession` to use for all
             commands
-          - `disable_md5` (optional): When True, an MD5 checksum will not be
-            computed for the uploaded file. Useful in environments where
-            MD5 cannot be used for regulatory or other reasons. Defaults to
-            False.
           - `**kwargs` (optional): file level options (see above)
 
+        .. versionchanged:: 4.0
+           Removed the `disable_md5` parameter. See
+           :ref:`removed-gridfs-checksum` for details.
+
+        .. versionchanged:: 3.7
+           Added the `disable_md5` parameter.
+
         .. versionchanged:: 3.6
            Added ``session`` parameter.
 
@@ -159,6 +173,7 @@ class GridIn(object):
         if not root_collection.write_concern.acknowledged:
             raise ConfigurationError('root_collection must use '
                                      'acknowledged write_concern')
+        _disallow_transactions(session)
 
         # Handle alternative naming
         if "content_type" in kwargs:
@@ -166,11 +181,9 @@ class GridIn(object):
         if "chunk_size" in kwargs:
             kwargs["chunkSize"] = kwargs.pop("chunk_size")
 
-        coll = root_collection.with_options(
-            read_preference=ReadPreference.PRIMARY)
+        coll = _clear_entity_type_registry(
+            root_collection, read_preference=ReadPreference.PRIMARY)
 
-        if not disable_md5:
-            kwargs["md5"] = hashlib.md5()
         # Defaults
         kwargs["_id"] = kwargs.get("_id", ObjectId())
         kwargs["chunkSize"] = kwargs.get("chunkSize", DEFAULT_CHUNK_SIZE)
@@ -178,7 +191,7 @@ class GridIn(object):
         object.__setattr__(self, "_coll", coll)
         object.__setattr__(self, "_chunks", coll.chunks)
         object.__setattr__(self, "_file", kwargs)
-        object.__setattr__(self, "_buffer", StringIO())
+        object.__setattr__(self, "_buffer", io.BytesIO())
         object.__setattr__(self, "_position", 0)
         object.__setattr__(self, "_chunk_number", 0)
         object.__setattr__(self, "_closed", False)
@@ -198,6 +211,7 @@ class GridIn(object):
 
     def __ensure_indexes(self):
         if not object.__getattribute__(self, "_ensured_index"):
+            _disallow_transactions(self._session)
             self.__create_index(self._coll.files, _F_INDEX, False)
             self.__create_index(self._coll.chunks, _C_INDEX, True)
             object.__setattr__(self, "_ensured_index", True)
@@ -256,9 +270,6 @@ class GridIn(object):
         """Flush `data` to a chunk.
         """
         self.__ensure_indexes()
-        if 'md5' in self._file:
-            self._file['md5'].update(data)
-
         if not data:
             return
         assert(len(data) <= self.chunk_size)
@@ -279,17 +290,15 @@ class GridIn(object):
         """
         self.__flush_data(self._buffer.getvalue())
         self._buffer.close()
-        self._buffer = StringIO()
+        self._buffer = io.BytesIO()
 
     def __flush(self):
         """Flush the file to the database.
         """
         try:
             self.__flush_buffer()
-
-            if "md5" in self._file:
-                self._file["md5"] = self._file["md5"].hexdigest()
-            self._file["length"] = self._position
+            # The GridFS spec says length SHOULD be an Int64.
+            self._file["length"] = Int64(self._position)
             self._file["uploadDate"] = datetime.datetime.utcnow()
 
             return self._coll.files.insert_one(
@@ -311,21 +320,29 @@ class GridIn(object):
             self.__flush()
             object.__setattr__(self, "_closed", True)
 
+    def read(self, size=-1):
+        raise io.UnsupportedOperation('read')
+
+    def readable(self):
+        return False
+
+    def seekable(self):
+        return False
+
     def write(self, data):
         """Write data to the file. There is no return value.
 
         `data` can be either a string of bytes or a file-like object
         (implementing :meth:`read`). If the file has an
         :attr:`encoding` attribute, `data` can also be a
-        :class:`unicode` (:class:`str` in python 3) instance, which
-        will be encoded as :attr:`encoding` before being written.
+        :class:`str` instance, which will be encoded as
+        :attr:`encoding` before being written.
 
         Due to buffering, the data may not actually be written to the
         database until the :meth:`close` method is called. Raises
         :class:`ValueError` if this file is already closed. Raises
         :class:`TypeError` if `data` is not an instance of
-        :class:`str` (:class:`bytes` in python 3), a file-like object,
-        or an instance of :class:`unicode` (:class:`str` in python 3).
+        :class:`bytes`, a file-like object, or an instance of :class:`str`.
         Unicode data is only allowed if the file has an :attr:`encoding`
         attribute.
 
@@ -341,15 +358,15 @@ class GridIn(object):
             read = data.read
         except AttributeError:
             # string
-            if not isinstance(data, (text_type, bytes)):
+            if not isinstance(data, (str, bytes)):
                 raise TypeError("can only write strings or file-like objects")
-            if isinstance(data, text_type):
+            if isinstance(data, str):
                 try:
                     data = data.encode(self.encoding)
                 except AttributeError:
                     raise TypeError("must specify an encoding for file in "
-                                    "order to write %s" % (text_type.__name__,))
-            read = StringIO(data).read
+                                    "order to write str")
+            read = io.BytesIO(data).read
 
         if self._buffer.tell() > 0:
             # Make sure to flush only when _buffer is complete
@@ -378,6 +395,9 @@ class GridIn(object):
         for line in sequence:
             self.write(line)
 
+    def writeable(self):
+        return True
+
     def __enter__(self):
         """Support for the context manager protocol.
         """
@@ -394,7 +414,7 @@ class GridIn(object):
         return False
 
 
-class GridOut(object):
+class GridOut(io.IOBase):
     """Class to read data out of GridFS.
     """
     def __init__(self, root_collection, file_id=None, file_document=None,
@@ -419,6 +439,11 @@ class GridOut(object):
             :class:`~pymongo.client_session.ClientSession` to use for all
             commands
 
+        .. versionchanged:: 3.8
+           For better performance and to better follow the GridFS spec,
+           :class:`GridOut` now uses a single cursor to read all the chunks in
+           the file.
+
         .. versionchanged:: 3.6
            Added ``session`` parameter.
 
@@ -429,11 +454,17 @@ class GridOut(object):
         if not isinstance(root_collection, Collection):
             raise TypeError("root_collection must be an "
                             "instance of Collection")
+        _disallow_transactions(session)
+
+        root_collection = _clear_entity_type_registry(root_collection)
+
+        super().__init__()
 
         self.__chunks = root_collection.chunks
         self.__files = root_collection.files
         self.__file_id = file_id
         self.__buffer = EMPTY
+        self.__chunk_iter = None
         self.__position = 0
         self._file = file_document
         self._session = session
@@ -453,6 +484,7 @@ class GridOut(object):
 
     def _ensure_file(self):
         if not self._file:
+            _disallow_transactions(self._session)
             self._file = self.__files.find_one({"_id": self.__file_id},
                                                session=self._session)
             if not self._file:
@@ -465,6 +497,9 @@ class GridOut(object):
             return self._file[name]
         raise AttributeError("GridOut object has no attribute '%s'" % name)
 
+    def readable(self):
+        return True
+
     def readchunk(self):
         """Reads a chunk at a time. If the current position is within a
         chunk the remainder of the chunk is returned.
@@ -477,12 +512,11 @@ class GridOut(object):
             chunk_data = self.__buffer
         elif self.__position < int(self.length):
             chunk_number = int((received + self.__position) / chunk_size)
-            chunk = self.__chunks.find_one({"files_id": self._id,
-                                            "n": chunk_number},
-                                           session=self._session)
-            if not chunk:
-                raise CorruptGridFile("no chunk #%d" % chunk_number)
+            if self.__chunk_iter is None:
+                self.__chunk_iter = _GridOutChunkIterator(
+                    self, self.__chunks, self._session, chunk_number)
 
+            chunk = self.__chunk_iter.next()
             chunk_data = chunk["data"][self.__position % chunk_size:]
 
             if not chunk_data:
@@ -501,33 +535,34 @@ class GridOut(object):
 
         :Parameters:
           - `size` (optional): the number of bytes to read
+
+        .. versionchanged:: 3.8
+           This method now only checks for extra chunks after reading the
+           entire file. Previously, this method would check for extra chunks
+           on every call.
         """
         self._ensure_file()
 
-        if size == 0:
-            return EMPTY
-
         remainder = int(self.length) - self.__position
         if size < 0 or size > remainder:
             size = remainder
 
+        if size == 0:
+            return EMPTY
+
         received = 0
-        data = StringIO()
+        data = io.BytesIO()
         while received < size:
             chunk_data = self.readchunk()
             received += len(chunk_data)
             data.write(chunk_data)
 
-        # Detect extra chunks.
-        max_chunk_n = math.ceil(self.length / float(self.chunk_size))
-        chunk = self.__chunks.find_one({"files_id": self._id,
-                                        "n": {"$gte": max_chunk_n}},
-                                       session=self._session)
-        # According to spec, ignore extra chunks if they are empty.
-        if chunk is not None and len(chunk['data']):
-            raise CorruptGridFile(
-                "Extra chunk found: expected %i chunks but found "
-                "chunk with n=%i" % (max_chunk_n, chunk['n']))
+        # Detect extra chunks after reading the entire file.
+        if size == remainder and self.__chunk_iter:
+            try:
+                self.__chunk_iter.next()
+            except StopIteration:
+                pass
 
         self.__position -= received - size
 
@@ -543,15 +578,15 @@ class GridOut(object):
         :Parameters:
          - `size` (optional): the maximum number of bytes to read
         """
-        if size == 0:
-            return b''
-
         remainder = int(self.length) - self.__position
         if size < 0 or size > remainder:
             size = remainder
 
+        if size == 0:
+            return EMPTY
+
         received = 0
-        data = StringIO()
+        data = io.BytesIO()
         while received < size:
             chunk_data = self.readchunk()
             pos = chunk_data.find(NEWLN, 0, size)
@@ -600,22 +635,55 @@ class GridOut(object):
         if new_pos < 0:
             raise IOError(22, "Invalid value for `pos` - must be positive")
 
+        # Optimization, continue using the same buffer and chunk iterator.
+        if new_pos == self.__position:
+            return
+
         self.__position = new_pos
         self.__buffer = EMPTY
+        if self.__chunk_iter:
+            self.__chunk_iter.close()
+            self.__chunk_iter = None
+
+    def seekable(self):
+        return True
 
     def __iter__(self):
         """Return an iterator over all of this file's data.
 
-        The iterator will return chunk-sized instances of
-        :class:`str` (:class:`bytes` in python 3). This can be
-        useful when serving files using a webserver that handles
-        such an iterator efficiently.
+        The iterator will return lines (delimited by ``b'\\n'``) of
+        :class:`bytes`. This can be useful when serving files
+        using a webserver that handles such an iterator efficiently.
+
+        .. versionchanged:: 3.8
+           The iterator now raises :class:`CorruptGridFile` when encountering
+           any truncated, missing, or extra chunk in a file. The previous
+           behavior was to only raise :class:`CorruptGridFile` on a missing
+           chunk.
+
+        .. versionchanged:: 4.0
+           The iterator now iterates over *lines* in the file, instead
+           of chunks, to conform to the base class :py:class:`io.IOBase`.
+           Use :meth:`GridOut.readchunk` to read chunk by chunk instead
+           of line by line.
         """
-        return GridOutIterator(self, self.__chunks, self._session)
+        return self
 
     def close(self):
         """Make GridOut more generically file-like."""
-        pass
+        if self.__chunk_iter:
+            self.__chunk_iter.close()
+            self.__chunk_iter = None
+        super().close()
+
+    def write(self, value):
+        raise io.UnsupportedOperation('write')
+
+    def writelines(self, lines):
+        raise io.UnsupportedOperation('writelines')
+
+    def writable(self):
+        return False
 
     def __enter__(self):
         """Makes it possible to use :class:`GridOut` files
@@ -627,30 +695,130 @@ class GridOut(object):
         """Makes it possible to use :class:`GridOut` files
         with the context manager protocol.
         """
+        self.close()
         return False
 
+    def fileno(self):
+        raise io.UnsupportedOperation('fileno')
+
+    def flush(self):
+        # GridOut is read-only, so flush does nothing.
+        pass
+
+    def isatty(self):
+        return False
+
+    def truncate(self, size=None):
+        # See https://docs.python.org/3/library/io.html#io.IOBase.writable
+        # for why truncate has to raise.
+        raise io.UnsupportedOperation('truncate')
+
+    # Override IOBase.__del__ otherwise it will lead to __getattr__ on
+    # __IOBase_closed which calls _ensure_file and potentially performs I/O.
+    # We cannot do I/O in __del__ since it can lead to a deadlock.
+    def __del__(self):
+        pass
+
+
+class _GridOutChunkIterator(object):
+    """Iterates over a file's chunks using a single cursor.
+
+    Raises CorruptGridFile when encountering any truncated, missing, or extra
+    chunk in a file.
+    """
+    def __init__(self, grid_out, chunks, session, next_chunk):
+        self._id = grid_out._id
+        self._chunk_size = int(grid_out.chunk_size)
+        self._length = int(grid_out.length)
+        self._chunks = chunks
+        self._session = session
+        self._next_chunk = next_chunk
+        self._num_chunks = math.ceil(float(self._length) / self._chunk_size)
+        self._cursor = None
+
+    def expected_chunk_length(self, chunk_n):
+        if chunk_n < self._num_chunks - 1:
+            return self._chunk_size
+        return self._length - (self._chunk_size * (self._num_chunks - 1))
+
+    def __iter__(self):
+        return self
+
+    def _create_cursor(self):
+        filter = {"files_id": self._id}
+        if self._next_chunk > 0:
+            filter["n"] = {"$gte": self._next_chunk}
+        _disallow_transactions(self._session)
+        self._cursor = self._chunks.find(filter, sort=[("n", 1)],
+                                         session=self._session)
+
+    def _next_with_retry(self):
+        """Return the next chunk and retry once on CursorNotFound.
+
+        We retry on CursorNotFound to maintain backwards compatibility in
+        cases where two calls to read occur more than 10 minutes apart (the
+        server's default cursor timeout).
+        """
+        if self._cursor is None:
+            self._create_cursor()
+
+        try:
+            return self._cursor.next()
+        except CursorNotFound:
+            self._cursor.close()
+            self._create_cursor()
+            return self._cursor.next()
+
+    def next(self):
+        try:
+            chunk = self._next_with_retry()
+        except StopIteration:
+            if self._next_chunk >= self._num_chunks:
+                raise
+            raise CorruptGridFile("no chunk #%d" % self._next_chunk)
+
+        if chunk["n"] != self._next_chunk:
+            self.close()
+            raise CorruptGridFile(
+                "Missing chunk: expected chunk #%d but found "
+                "chunk with n=%d" % (self._next_chunk, chunk["n"]))
+
+        if chunk["n"] >= self._num_chunks:
+            # According to spec, ignore extra chunks if they are empty.
+            if len(chunk["data"]):
+                self.close()
+                raise CorruptGridFile(
+                    "Extra chunk found: expected %d chunks but found "
+                    "chunk with n=%d" % (self._num_chunks, chunk["n"]))
+
+        expected_length = self.expected_chunk_length(chunk["n"])
+        if len(chunk["data"]) != expected_length:
+            self.close()
+            raise CorruptGridFile(
+                "truncated chunk #%d: expected chunk length to be %d but "
+                "found chunk with length %d" % (
+                    chunk["n"], expected_length, len(chunk["data"])))
+
+        self._next_chunk += 1
+        return chunk
+
+    __next__ = next
+
+    def close(self):
+        if self._cursor:
+            self._cursor.close()
+            self._cursor = None
+
 
 class GridOutIterator(object):
     def __init__(self, grid_out, chunks, session):
-        self.__id = grid_out._id
-        self.__chunks = chunks
-        self.__session = session
-        self.__current_chunk = 0
-        self.__max_chunk = math.ceil(float(grid_out.length) /
-                                     grid_out.chunk_size)
+        self.__chunk_iter = _GridOutChunkIterator(grid_out, chunks, session, 0)
 
     def __iter__(self):
         return self
 
     def next(self):
-        if self.__current_chunk >= self.__max_chunk:
-            raise StopIteration
-        chunk = self.__chunks.find_one({"files_id": self.__id,
-                                        "n": self.__current_chunk},
-                                       session=self.__session)
-        if not chunk:
-            raise CorruptGridFile("no chunk #%d" % self.__current_chunk)
-        self.__current_chunk += 1
+        chunk = self.__chunk_iter.next()
         return bytes(chunk["data"])
 
     __next__ = next
@@ -671,8 +839,11 @@ class GridOutCursor(Cursor):
 
         .. versionadded 2.7
 
-        .. mongodoc:: cursors
+        .. seealso:: The MongoDB documentation on `cursors <https://dochub.mongodb.org/core/cursors>`_.
         """
+        _disallow_transactions(session)
+        collection = _clear_entity_type_registry(collection)
+
         # Hold on to the base "fs" collection to create GridOut objects later.
         self.__root_collection = collection
 
@@ -684,6 +855,7 @@ class GridOutCursor(Cursor):
     def next(self):
         """Get next GridOut object from cursor.
         """
+        _disallow_transactions(self.session)
         # Work around "super is not iterable" issue in Python 3.x
         next_file = super(GridOutCursor, self).next()
         return GridOut(self.__root_collection, file_document=next_file,
