@@ -15,35 +15,41 @@
 """Internal class to monitor a topology of one or more servers."""
 
 import os
+import queue
 import random
 import threading
+import time
 import warnings
 import weakref
 
-from bson.py3compat import itervalues, PY3
-if PY3:
-    import queue as Queue
-else:
-    import Queue
-
-from pymongo import common
-from pymongo import periodic_executor
-from pymongo.pool import PoolOptions
-from pymongo.topology_description import (updated_topology_description,
-                                          _updated_topology_description_srv_polling,
-                                          TopologyDescription,
-                                          SRV_POLLING_TOPOLOGIES, TOPOLOGY_TYPE)
-from pymongo.errors import ServerSelectionTimeoutError, ConfigurationError
+from pymongo import (common,
+                     helpers,
+                     periodic_executor)
+from pymongo.client_session import _ServerSessionPool
+from pymongo.errors import (ConnectionFailure,
+                            ConfigurationError,
+                            NetworkTimeout,
+                            NotPrimaryError,
+                            OperationFailure,
+                            PyMongoError,
+                            ServerSelectionTimeoutError,
+                            WriteError,
+                            InvalidOperation)
+from pymongo.hello import Hello
 from pymongo.monitor import SrvMonitor
-from pymongo.monotonic import time as _time
+from pymongo.pool import PoolOptions
 from pymongo.server import Server
+from pymongo.server_description import ServerDescription
 from pymongo.server_selectors import (any_server_selector,
                                       arbiter_server_selector,
                                       secondary_server_selector,
                                       readable_server_selector,
                                       writable_server_selector,
                                       Selection)
-from pymongo.client_session import _ServerSessionPool
+from pymongo.topology_description import (updated_topology_description,
+                                          _updated_topology_description_srv_polling,
+                                          TopologyDescription,
+                                          SRV_POLLING_TOPOLOGIES, TOPOLOGY_TYPE)
 
 
 def process_events_queue(queue_ref):
@@ -54,7 +60,7 @@ def process_events_queue(queue_ref):
     while True:
         try:
             event = q.get_nowait()
-        except Queue.Empty:
+        except queue.Empty:
             break
         else:
             fn, args = event
@@ -67,7 +73,7 @@ class Topology(object):
     """Monitor a topology of one or more servers."""
     def __init__(self, topology_settings):
         self._topology_id = topology_settings._topology_id
-        self._listeners = topology_settings._pool_options.event_listeners
+        self._listeners = topology_settings._pool_options._event_listeners
         pub = self._listeners is not None
         self._publish_server = pub and self._listeners.enabled_for_server
         self._publish_tp = pub and self._listeners.enabled_for_topology
@@ -77,7 +83,7 @@ class Topology(object):
         self.__events_executor = None
 
         if self._publish_server or self._publish_tp:
-            self._events = Queue.Queue(maxsize=100)
+            self._events = queue.Queue(maxsize=100)
 
         if self._publish_tp:
             self._events.put((self._listeners.publish_topology_opened,
@@ -107,6 +113,7 @@ class Topology(object):
         # Store the seed list to help diagnose errors in _error_message().
         self._seed_addresses = list(topology_description.server_descriptions())
         self._opened = False
+        self._closed = False
         self._lock = threading.Lock()
         self._condition = self._settings.condition_class(self._lock)
         self._servers = {}
@@ -120,7 +127,7 @@ class Topology(object):
 
             executor = periodic_executor.PeriodicExecutor(
                 interval=common.EVENTS_QUEUE_FREQUENCY,
-                min_interval=0.5,
+                min_interval=common.MIN_HEARTBEAT_INTERVAL,
                 target=target,
                 name="pymongo_events_thread")
 
@@ -132,7 +139,8 @@ class Topology(object):
             executor.open()
 
         self._srv_monitor = None
-        if self._settings.fqdn is not None:
+        if (self._settings.fqdn is not None and
+                not self._settings.load_balanced):
             self._srv_monitor = SrvMonitor(self, self._settings)
 
     def open(self):
@@ -154,7 +162,7 @@ class Topology(object):
                 warnings.warn(
                     "MongoClient opened before fork. Create MongoClient only "
                     "after forking. See PyMongo's documentation for details: "
-                    "http://api.mongodb.org/python/current/faq.html#"
+                    "https://pymongo.readthedocs.io/en/stable/faq.html#"
                     "is-pymongo-fork-safe")
                 with self._lock:
                     # Reset the session pool to avoid duplicate sessions in
@@ -197,7 +205,7 @@ class Topology(object):
 
     def _select_servers_loop(self, selector, timeout, address):
         """select_servers() guts. Hold the lock when calling this."""
-        now = _time()
+        now = time.monotonic()
         end_time = now + timeout
         server_descriptions = self._description.apply_selector(
             selector, address, custom_selector=self._settings.server_selector)
@@ -206,7 +214,8 @@ class Topology(object):
             # No suitable servers.
             if timeout == 0 or now > end_time:
                 raise ServerSelectionTimeoutError(
-                    self._error_message(selector))
+                    "%s, Timeout: %ss, Topology Description: %r" %
+                    (self._error_message(selector), timeout, self.description))
 
             self._ensure_opened()
             self._request_check_all()
@@ -217,7 +226,7 @@ class Topology(object):
             # held the lock until now.
             self._condition.wait(common.MIN_HEARTBEAT_INTERVAL)
             self._description.check_compatible()
-            now = _time()
+            now = time.monotonic()
             server_descriptions = self._description.apply_selector(
                 selector, address,
                 custom_selector=self._settings.server_selector)
@@ -230,9 +239,15 @@ class Topology(object):
                       server_selection_timeout=None,
                       address=None):
         """Like select_servers, but choose a random server if several match."""
-        return random.choice(self.select_servers(selector,
-                                                 server_selection_timeout,
-                                                 address))
+        servers = self.select_servers(
+            selector, server_selection_timeout, address)
+        if len(servers) == 1:
+            return servers[0]
+        server1, server2 = random.sample(servers, 2)
+        if server1.pool.operation_count <= server2.pool.operation_count:
+            return server1
+        else:
+            return server2
 
     def select_server_by_address(self, address,
                                  server_selection_timeout=None):
@@ -257,27 +272,40 @@ class Topology(object):
                                   server_selection_timeout,
                                   address)
 
-    def _process_change(self, server_description):
+    def _process_change(self, server_description, reset_pool=False):
         """Process a new ServerDescription on an opened topology.
 
         Hold the lock when calling this.
         """
         td_old = self._description
-        if self._publish_server:
-            old_server_description = td_old._server_descriptions[
-                server_description.address]
+        sd_old = td_old._server_descriptions[server_description.address]
+        if _is_stale_server_description(sd_old, server_description):
+            # This is a stale hello response. Ignore it.
+            return
+
+        new_td = updated_topology_description(
+            self._description, server_description)
+        # CMAP: Ensure the pool is "ready" when the server is selectable.
+        if (server_description.is_readable
+                or (server_description.is_server_type_known and
+                    new_td.topology_type == TOPOLOGY_TYPE.Single)):
+            server = self._servers.get(server_description.address)
+            if server:
+                server.pool.ready()
+
+        suppress_event = ((self._publish_server or self._publish_tp)
+                          and sd_old == server_description)
+        if self._publish_server and not suppress_event:
             self._events.put((
                 self._listeners.publish_server_description_changed,
-                (old_server_description, server_description,
+                (sd_old, server_description,
                  server_description.address, self._topology_id)))
 
-        self._description = updated_topology_description(
-            self._description, server_description)
-
+        self._description = new_td
         self._update_servers()
         self._receive_cluster_time_no_lock(server_description.cluster_time)
 
-        if self._publish_tp:
+        if self._publish_tp and not suppress_event:
             self._events.put((
                 self._listeners.publish_topology_description_changed,
                 (td_old, self._description, self._topology_id)))
@@ -290,14 +318,20 @@ class Topology(object):
                                   SRV_POLLING_TOPOLOGIES):
             self._srv_monitor.close()
 
+        # Clear the pool from a failed heartbeat.
+        if reset_pool:
+            server = self._servers.get(server_description.address)
+            if server:
+                server.pool.reset()
+
         # Wake waiters in select_servers().
         self._condition.notify_all()
 
-    def on_change(self, server_description):
-        """Process a new ServerDescription after an ismaster call completes."""
+    def on_change(self, server_description, reset_pool=False):
+        """Process a new ServerDescription after an hello call completes."""
         # We do no I/O holding the lock.
         with self._lock:
-            # Monitors may continue working on ismaster calls for some time
+            # Monitors may continue working on hello calls for some time
             # after a call to Topology.close, so this method may be called at
             # any time. Ensure the topology is open before processing the
             # change.
@@ -307,7 +341,7 @@ class Topology(object):
             # that didn't include this server.
             if (self._opened and
                     self._description.has_server(server_description.address)):
-                self._process_change(server_description)
+                self._process_change(server_description, reset_pool)
 
     def _process_srv_update(self, seedlist):
         """Process a new seedlist on an opened topology.
@@ -401,44 +435,37 @@ class Topology(object):
             self._request_check_all()
             self._condition.wait(wait_time)
 
-    def reset_pool(self, address):
-        with self._lock:
-            server = self._servers.get(address)
-            if server:
-                server.pool.reset()
+    def data_bearing_servers(self):
+        """Return a list of all data-bearing servers.
 
-    def reset_server(self, address):
-        """Clear our pool for a server and mark it Unknown.
-
-        Do *not* request an immediate check.
+        This includes any server that might be selected for an operation.
         """
-        with self._lock:
-            self._reset_server(address, reset_pool=True)
+        if self._description.topology_type == TOPOLOGY_TYPE.Single:
+            return self._description.known_servers
+        return self._description.readable_servers
 
-    def reset_server_and_request_check(self, address):
-        """Clear our pool for a server, mark it Unknown, and check it soon."""
-        with self._lock:
-            self._reset_server(address, reset_pool=True)
-            self._request_check(address)
-
-    def mark_server_unknown_and_request_check(self, address):
-        """Mark a server Unknown, and check it soon."""
-        with self._lock:
-            self._reset_server(address, reset_pool=False)
-            self._request_check(address)
-
-    def update_pool(self):
+    def update_pool(self, all_credentials):
         # Remove any stale sockets and add new sockets if pool is too small.
         servers = []
         with self._lock:
-            for server in self._servers.values():
-                servers.append((server, server._pool.pool_id))
+            # Only update pools for data-bearing servers.
+            for sd in self.data_bearing_servers():
+                server = self._servers[sd.address]
+                servers.append((server,
+                                server.pool.gen.get_overall()))
 
-        for server, pool_id in servers:
-            server._pool.remove_stale_sockets(pool_id)
+        for server, generation in servers:
+            try:
+                server.pool.remove_stale_sockets(generation, all_credentials)
+            except PyMongoError as exc:
+                ctx = _ErrorContext(exc, 0, generation, False, None)
+                self.handle_error(server.description.address, ctx)
+                raise
 
     def close(self):
-        """Clear pools and terminate monitors. Topology reopens on demand."""
+        """Clear pools and terminate monitors. Topology does not reopen on
+        demand. Any further operations will raise
+        :exc:`~.errors.InvalidOperation`. """
         with self._lock:
             for server in self._servers.values():
                 server.close()
@@ -454,6 +481,7 @@ class Topology(object):
                 self._srv_monitor.close()
 
             self._opened = False
+            self._closed = True
 
         # Publish only after releasing the lock.
         if self._publish_tp:
@@ -471,39 +499,46 @@ class Topology(object):
         with self._lock:
             return self._session_pool.pop_all()
 
-    def get_server_session(self):
-        """Start or resume a server session, or raise ConfigurationError."""
-        with self._lock:
-            session_timeout = self._description.logical_session_timeout_minutes
-            if session_timeout is None:
-                # Maybe we need an initial scan? Can raise ServerSelectionError.
-                if self._description.topology_type == TOPOLOGY_TYPE.Single:
-                    if not self._description.has_known_servers:
-                        self._select_servers_loop(
-                            any_server_selector,
-                            self._settings.server_selection_timeout,
-                            None)
-                elif not self._description.readable_servers:
+    def _check_session_support(self):
+        """Internal check for session support on non-load balanced clusters."""
+        session_timeout = self._description.logical_session_timeout_minutes
+        if session_timeout is None:
+            # Maybe we need an initial scan? Can raise ServerSelectionError.
+            if self._description.topology_type == TOPOLOGY_TYPE.Single:
+                if not self._description.has_known_servers:
                     self._select_servers_loop(
-                        readable_server_selector,
+                        any_server_selector,
                         self._settings.server_selection_timeout,
                         None)
+            elif not self._description.readable_servers:
+                self._select_servers_loop(
+                    readable_server_selector,
+                    self._settings.server_selection_timeout,
+                    None)
 
             session_timeout = self._description.logical_session_timeout_minutes
             if session_timeout is None:
                 raise ConfigurationError(
                     "Sessions are not supported by this MongoDB deployment")
+        return session_timeout
 
+    def get_server_session(self):
+        """Start or resume a server session, or raise ConfigurationError."""
+        with self._lock:
+            # Sessions are always supported in load balanced mode.
+            if not self._settings.load_balanced:
+                session_timeout = self._check_session_support()
+            else:
+                # Sessions never time out in load balanced mode.
+                session_timeout = float('inf')
             return self._session_pool.get_server_session(session_timeout)
 
     def return_server_session(self, server_session, lock):
         if lock:
             with self._lock:
-                session_timeout = \
-                    self._description.logical_session_timeout_minutes
-                if session_timeout is not None:
-                    self._session_pool.return_server_session(server_session,
-                                                             session_timeout)
+                self._session_pool.return_server_session(
+                    server_session,
+                    self._description.logical_session_timeout_minutes)
         else:
             # Called from a __del__ method, can't use a lock.
             self._session_pool.return_server_session_no_lock(server_session)
@@ -520,6 +555,9 @@ class Topology(object):
 
         Hold the lock when calling this.
         """
+        if self._closed:
+            raise InvalidOperation("Cannot use MongoClient after close")
+
         if not self._opened:
             self._opened = True
             self._update_servers()
@@ -533,33 +571,106 @@ class Topology(object):
                                       SRV_POLLING_TOPOLOGIES):
                 self._srv_monitor.open()
 
+            if self._settings.load_balanced:
+                # Emit initial SDAM events for load balancer mode.
+                self._process_change(ServerDescription(
+                    self._seed_addresses[0],
+                    Hello({'ok': 1, 'serviceId': self._topology_id,
+                              'maxWireVersion': 13})))
+
         # Ensure that the monitors are open.
-        for server in itervalues(self._servers):
+        for server in self._servers.values():
             server.open()
 
-    def _reset_server(self, address, reset_pool):
-        """Mark a server Unknown and optionally reset it's pool.
+    def _is_stale_error(self, address, err_ctx):
+        server = self._servers.get(address)
+        if server is None:
+            # Another thread removed this server from the topology.
+            return True
 
-        Hold the lock when calling this. Does *not* request an immediate check.
+        if server._pool.stale_generation(
+                err_ctx.sock_generation, err_ctx.service_id):
+            # This is an outdated error from a previous pool version.
+            return True
+
+        # topologyVersion check, ignore error when cur_tv >= error_tv:
+        cur_tv = server.description.topology_version
+        error = err_ctx.error
+        error_tv = None
+        if error and hasattr(error, 'details'):
+            if isinstance(error.details, dict):
+                error_tv = error.details.get('topologyVersion')
+
+        return _is_stale_error_topology_version(cur_tv, error_tv)
+
+    def _handle_error(self, address, err_ctx):
+        if self._is_stale_error(address, err_ctx):
+            return
+
+        server = self._servers[address]
+        error = err_ctx.error
+        exc_type = type(error)
+        service_id = err_ctx.service_id
+        if (issubclass(exc_type, NetworkTimeout) and
+                err_ctx.completed_handshake):
+            # The socket has been closed. Don't reset the server.
+            # Server Discovery And Monitoring Spec: "When an application
+            # operation fails because of any network error besides a socket
+            # timeout...."
+            return
+        elif issubclass(exc_type, WriteError):
+            # Ignore writeErrors.
+            return
+        elif issubclass(exc_type, (NotPrimaryError, OperationFailure)):
+            # As per the SDAM spec if:
+            #   - the server sees a "not primary" error, and
+            #   - the server is not shutting down, and
+            #   - the server version is >= 4.2, then
+            # we keep the existing connection pool, but mark the server type
+            # as Unknown and request an immediate check of the server.
+            # Otherwise, we clear the connection pool, mark the server as
+            # Unknown and request an immediate check of the server.
+            if hasattr(error, 'code'):
+                err_code = error.code
+            else:
+                # Default error code if one does not exist.
+                default = 10107 if isinstance(error, NotPrimaryError) else None
+                err_code = error.details.get('code', default)
+            if err_code in helpers._NOT_PRIMARY_CODES:
+                is_shutting_down = err_code in helpers._SHUTDOWN_CODES
+                # Mark server Unknown, clear the pool, and request check.
+                if not self._settings.load_balanced:
+                    self._process_change(ServerDescription(address, error=error))
+                if is_shutting_down or (err_ctx.max_wire_version <= 7):
+                    # Clear the pool.
+                    server.reset(service_id)
+                server.request_check()
+            elif not err_ctx.completed_handshake:
+                # Unknown command error during the connection handshake.
+                if not self._settings.load_balanced:
+                    self._process_change(ServerDescription(address, error=error))
+                # Clear the pool.
+                server.reset(service_id)
+        elif issubclass(exc_type, ConnectionFailure):
+            # "Client MUST replace the server's description with type Unknown
+            # ... MUST NOT request an immediate check of the server."
+            if not self._settings.load_balanced:
+                self._process_change(ServerDescription(address, error=error))
+            # Clear the pool.
+            server.reset(service_id)
+            # "When a client marks a server Unknown from `Network error when
+            # reading or writing`_, clients MUST cancel the hello check on
+            # that server and close the current monitoring connection."
+            server._monitor.cancel_check()
+
+    def handle_error(self, address, err_ctx):
+        """Handle an application error.
+
+        May reset the server to Unknown, clear the pool, and request an
+        immediate check depending on the error and the context.
         """
-        server = self._servers.get(address)
-
-        # "server" is None if another thread removed it from the topology.
-        if server:
-            if reset_pool:
-                server.reset()
-
-            # Mark this server Unknown.
-            self._description = self._description.reset_server(address)
-            self._update_servers()
-
-    def _request_check(self, address):
-        """Wake one monitor. Hold the lock when calling this."""
-        server = self._servers.get(address)
-
-        # "server" is None if another thread removed it from the topology.
-        if server:
-            server.request_check()
+        with self._lock:
+            self._handle_error(address, err_ctx)
 
     def _request_check_all(self):
         """Wake all monitors. Hold the lock when calling this."""
@@ -619,11 +730,14 @@ class Topology(object):
         monitor_pool_options = PoolOptions(
             connect_timeout=options.connect_timeout,
             socket_timeout=options.connect_timeout,
-            ssl_context=options.ssl_context,
-            ssl_match_hostname=options.ssl_match_hostname,
-            event_listeners=options.event_listeners,
+            ssl_context=options._ssl_context,
+            tls_allow_invalid_hostnames=options.tls_allow_invalid_hostnames,
+            event_listeners=options._event_listeners,
             appname=options.appname,
-            driver=options.driver)
+            driver=options.driver,
+            pause_enabled=False,
+            server_api=options.server_api,
+        )
 
         return self._settings.pool_class(address, monitor_pool_options,
                                          handshake=False)
@@ -684,3 +798,53 @@ class Topology(object):
             else:
                 return ','.join(str(server.error) for server in servers
                                 if server.error)
+
+    def __repr__(self):
+        msg = ''
+        if not self._opened:
+            msg = 'CLOSED '
+        return '<%s %s%r>' % (self.__class__.__name__, msg, self._description)
+
+    def eq_props(self):
+        """The properties to use for MongoClient/Topology equality checks."""
+        ts = self._settings
+        return (tuple(sorted(ts.seeds)), ts.replica_set_name, ts.fqdn,
+                ts.srv_service_name)
+
+    def __eq__(self, other):
+        if isinstance(other, self.__class__):
+            return self.eq_props() == other.eq_props()
+        return NotImplemented
+
+    def __hash__(self):
+        return hash(self.eq_props())
+
+
+class _ErrorContext(object):
+    """An error with context for SDAM error handling."""
+    def __init__(self, error, max_wire_version, sock_generation,
+                 completed_handshake, service_id):
+        self.error = error
+        self.max_wire_version = max_wire_version
+        self.sock_generation = sock_generation
+        self.completed_handshake = completed_handshake
+        self.service_id = service_id
+
+
+def _is_stale_error_topology_version(current_tv, error_tv):
+    """Return True if the error's topologyVersion is <= current."""
+    if current_tv is None or error_tv is None:
+        return False
+    if current_tv['processId'] != error_tv['processId']:
+        return False
+    return current_tv['counter'] >= error_tv['counter']
+
+
+def _is_stale_server_description(current_sd, new_sd):
+    """Return True if the new topologyVersion is < current."""
+    current_tv, new_tv = current_sd.topology_version, new_sd.topology_version
+    if current_tv is None or new_tv is None:
+        return False
+    if current_tv['processId'] != new_tv['processId']:
+        return False
+    return current_tv['counter'] > new_tv['counter']
