@@ -15,6 +15,7 @@
 """Represent a deployment of MongoDB servers."""
 
 from collections import namedtuple
+from random import sample
 
 from pymongo import common
 from pymongo.errors import ConfigurationError
@@ -25,9 +26,9 @@ from pymongo.server_type import SERVER_TYPE
 
 
 # Enumeration for various kinds of MongoDB cluster topologies.
-TOPOLOGY_TYPE = namedtuple('TopologyType', ['Single', 'ReplicaSetNoPrimary',
-                                            'ReplicaSetWithPrimary', 'Sharded',
-                                            'Unknown'])(*range(5))
+TOPOLOGY_TYPE = namedtuple('TopologyType', [
+    'Single', 'ReplicaSetNoPrimary', 'ReplicaSetWithPrimary', 'Sharded',
+    'Unknown', 'LoadBalanced'])(*range(6))
 
 # Topologies compatible with SRV record polling.
 SRV_POLLING_TOPOLOGIES = (TOPOLOGY_TYPE.Unknown, TOPOLOGY_TYPE.Sharded)
@@ -63,7 +64,28 @@ class TopologyDescription(object):
 
         # Is PyMongo compatible with all servers' wire protocols?
         self._incompatible_err = None
+        if self._topology_type != TOPOLOGY_TYPE.LoadBalanced:
+            self._init_incompatible_err()
 
+        # Server Discovery And Monitoring Spec: Whenever a client updates the
+        # TopologyDescription from an hello response, it MUST set
+        # TopologyDescription.logicalSessionTimeoutMinutes to the smallest
+        # logicalSessionTimeoutMinutes value among ServerDescriptions of all
+        # data-bearing server types. If any have a null
+        # logicalSessionTimeoutMinutes, then
+        # TopologyDescription.logicalSessionTimeoutMinutes MUST be set to null.
+        readable_servers = self.readable_servers
+        if not readable_servers:
+            self._ls_timeout_minutes = None
+        elif any(s.logical_session_timeout_minutes is None
+                 for s in readable_servers):
+            self._ls_timeout_minutes = None
+        else:
+            self._ls_timeout_minutes = min(s.logical_session_timeout_minutes
+                                           for s in readable_servers)
+
+    def _init_incompatible_err(self):
+        """Internal compatibility check for non-load balanced topologies."""
         for s in self._server_descriptions.values():
             if not s.is_server_type_known:
                 continue
@@ -98,23 +120,6 @@ class TopologyDescription(object):
 
                 break
 
-        # Server Discovery And Monitoring Spec: Whenever a client updates the
-        # TopologyDescription from an ismaster response, it MUST set
-        # TopologyDescription.logicalSessionTimeoutMinutes to the smallest
-        # logicalSessionTimeoutMinutes value among ServerDescriptions of all
-        # data-bearing server types. If any have a null
-        # logicalSessionTimeoutMinutes, then
-        # TopologyDescription.logicalSessionTimeoutMinutes MUST be set to null.
-        readable_servers = self.readable_servers
-        if not readable_servers:
-            self._ls_timeout_minutes = None
-        elif any(s.logical_session_timeout_minutes is None
-                 for s in readable_servers):
-            self._ls_timeout_minutes = None
-        else:
-            self._ls_timeout_minutes = min(s.logical_session_timeout_minutes
-                                           for s in readable_servers)
-
     def check_compatible(self):
         """Raise ConfigurationError if any server is incompatible.
 
@@ -129,7 +134,8 @@ class TopologyDescription(object):
 
     def reset_server(self, address):
         """A copy of this description, with one server marked Unknown."""
-        return updated_topology_description(self, ServerDescription(address))
+        unknown_sd = self._server_descriptions[address].to_unknown()
+        return updated_topology_description(self, unknown_sd)
 
     def reset(self):
         """A copy of this description, with all servers marked Unknown."""
@@ -218,21 +224,36 @@ class TopologyDescription(object):
     def heartbeat_frequency(self):
         return self._topology_settings.heartbeat_frequency
 
-    def apply_selector(self, selector, address, custom_selector=None):
+    @property
+    def srv_max_hosts(self):
+        return self._topology_settings._srv_max_hosts
 
-        def apply_local_threshold(selection):
-            if not selection:
-                return []
+    def _apply_local_threshold(self, selection):
+        if not selection:
+            return []
+        # Round trip time in seconds.
+        fastest = min(
+            s.round_trip_time for s in selection.server_descriptions)
+        threshold = self._topology_settings.local_threshold_ms / 1000.0
+        return [s for s in selection.server_descriptions
+                if (s.round_trip_time - fastest) <= threshold]
 
-            settings = self._topology_settings
+    def apply_selector(self, selector, address=None, custom_selector=None):
+        """List of servers matching the provided selector(s).
 
-            # Round trip time in seconds.
-            fastest = min(
-                s.round_trip_time for s in selection.server_descriptions)
-            threshold = settings.local_threshold_ms / 1000.0
-            return [s for s in selection.server_descriptions
-                    if (s.round_trip_time - fastest) <= threshold]
+        :Parameters:
+          - `selector`: a callable that takes a Selection as input and returns
+            a Selection as output. For example, an instance of a read
+            preference from :mod:`~pymongo.read_preferences`.
+          - `address` (optional): A server address to select.
+          - `custom_selector` (optional): A callable that augments server
+            selection rules. Accepts a list of
+            :class:`~pymongo.server_description.ServerDescription` objects and
+            return a list of server descriptions that should be considered
+            suitable for the desired operation.
 
+        .. versionadded:: 3.4
+        """
         if getattr(selector, 'min_wire_version', 0):
             common_wv = self.common_wire_version
             if common_wv and common_wv < selector.min_wire_version:
@@ -242,8 +263,10 @@ class TopologyDescription(object):
                                              selector.min_wire_version,
                                              common_wv))
 
-        if self.topology_type == TOPOLOGY_TYPE.Single:
-            # Ignore selectors for standalone.
+        if self.topology_type in (TOPOLOGY_TYPE.Single,
+                                  TOPOLOGY_TYPE.LoadBalanced,
+                                  TOPOLOGY_TYPE.Unknown):
+            # Ignore selectors for standalone and load balancer mode.
             return self.known_servers
         elif address:
             # Ignore selectors when explicit address is requested.
@@ -259,7 +282,7 @@ class TopologyDescription(object):
         if custom_selector is not None and selection:
             selection = selection.with_server_descriptions(
                 custom_selector(selection.server_descriptions))
-        return apply_local_threshold(selection)
+        return self._apply_local_threshold(selection)
 
     def has_readable_server(self, read_preference=ReadPreference.PRIMARY):
         """Does this topology have any readable servers available matching the
@@ -276,7 +299,7 @@ class TopologyDescription(object):
         .. versionadded:: 3.4
         """
         common.validate_read_preference("read_preference", read_preference)
-        return any(self.apply_selector(read_preference, None))
+        return any(self.apply_selector(read_preference))
 
     def has_writable_server(self):
         """Does this topology have a writable server available?
@@ -288,8 +311,16 @@ class TopologyDescription(object):
         """
         return self.has_readable_server(ReadPreference.PRIMARY)
 
+    def __repr__(self):
+        # Sort the servers by address.
+        servers = sorted(self._server_descriptions.values(),
+                         key=lambda sd: sd.address)
+        return "<%s id: %s, topology_type: %s, servers: %r>" % (
+            self.__class__.__name__, self._topology_settings._topology_id,
+            self.topology_type_name, servers)
 
-# If topology type is Unknown and we receive an ismaster response, what should
+
+# If topology type is Unknown and we receive a hello response, what should
 # the new topology type be?
 _SERVER_TYPE_TO_TOPOLOGY_TYPE = {
     SERVER_TYPE.Mongos: TOPOLOGY_TYPE.Sharded,
@@ -297,6 +328,7 @@ _SERVER_TYPE_TO_TOPOLOGY_TYPE = {
     SERVER_TYPE.RSSecondary: TOPOLOGY_TYPE.ReplicaSetNoPrimary,
     SERVER_TYPE.RSArbiter: TOPOLOGY_TYPE.ReplicaSetNoPrimary,
     SERVER_TYPE.RSOther: TOPOLOGY_TYPE.ReplicaSetNoPrimary,
+    # Note: SERVER_TYPE.LoadBalancer and Unknown are intentionally left out.
 }
 
 
@@ -306,9 +338,9 @@ def updated_topology_description(topology_description, server_description):
     :Parameters:
       - `topology_description`: the current TopologyDescription
       - `server_description`: a new ServerDescription that resulted from
-        an ismaster call
+        a hello call
 
-    Called after attempting (successfully or not) to call ismaster on the
+    Called after attempting (successfully or not) to call hello on the
     server at server_description.address. Does not modify topology_description.
     """
     address = server_description.address
@@ -328,6 +360,14 @@ def updated_topology_description(topology_description, server_description):
     sds[address] = server_description
 
     if topology_type == TOPOLOGY_TYPE.Single:
+        # Set server type to Unknown if replica set name does not match.
+        if (set_name is not None and
+                set_name != server_description.replica_set_name):
+            error = ConfigurationError(
+                "client is configured to connect to a replica set named "
+                "'%s' but this node belongs to a set named '%s'" % (
+                    set_name, server_description.replica_set_name))
+            sds[address] = server_description.to_unknown(error=error)
         # Single type never changes.
         return TopologyDescription(
             TOPOLOGY_TYPE.Single,
@@ -338,9 +378,12 @@ def updated_topology_description(topology_description, server_description):
             topology_description._topology_settings)
 
     if topology_type == TOPOLOGY_TYPE.Unknown:
-        if server_type == SERVER_TYPE.Standalone:
-            sds.pop(address)
-
+        if server_type in (SERVER_TYPE.Standalone, SERVER_TYPE.LoadBalancer):
+            if len(topology_description._topology_settings.seeds) == 1:
+                topology_type = TOPOLOGY_TYPE.Single
+            else:
+                # Remove standalone from Topology when given multiple seeds.
+                sds.pop(address)
         elif server_type not in (SERVER_TYPE.Unknown, SERVER_TYPE.RSGhost):
             topology_type = _SERVER_TYPE_TO_TOPOLOGY_TYPE[server_type]
 
@@ -410,7 +453,7 @@ def _updated_topology_description_srv_polling(topology_description, seedlist):
     :Parameters:
       - `topology_description`: the current TopologyDescription
       - `seedlist`: a list of new seeds new ServerDescription that resulted from
-        an ismaster call
+        a hello call
     """
     # Create a copy of the server descriptions.
     sds = topology_description.server_descriptions()
@@ -419,16 +462,23 @@ def _updated_topology_description_srv_polling(topology_description, seedlist):
     if set(sds.keys()) == set(seedlist):
         return topology_description
 
-    # Add SDs corresponding to servers recently added to the SRV record.
-    for address in seedlist:
-        if address not in sds:
-            sds[address] = ServerDescription(address)
 
     # Remove SDs corresponding to servers no longer part of the SRV record.
     for address in list(sds.keys()):
         if address not in seedlist:
             sds.pop(address)
 
+    if topology_description.srv_max_hosts != 0:
+        new_hosts = set(seedlist) - set(sds.keys())
+        n_to_add = topology_description.srv_max_hosts - len(sds)
+        if n_to_add > 0:
+            seedlist = sample(new_hosts, min(n_to_add, len(new_hosts)))
+        else:
+            seedlist = []
+    # Add SDs corresponding to servers recently added to the SRV record.
+    for address in seedlist:
+        if address not in sds:
+            sds[address] = ServerDescription(address)
     return TopologyDescription(
         topology_description.topology_type,
         sds,
@@ -444,7 +494,7 @@ def _update_rs_from_primary(
         server_description,
         max_set_version,
         max_election_id):
-    """Update topology description from a primary's ismaster response.
+    """Update topology description from a primary's hello response.
 
     Pass in a dict of ServerDescriptions, current replica set name, the
     ServerDescription we are processing, and the TopologyDescription's
@@ -471,8 +521,7 @@ def _update_rs_from_primary(
                 max_election_tuple > server_description.election_tuple):
 
             # Stale primary, set to type Unknown.
-            address = server_description.address
-            sds[address] = ServerDescription(address)
+            sds[server_description.address] = server_description.to_unknown()
             return (_check_has_primary(sds),
                     replica_set_name,
                     max_set_version,
@@ -492,7 +541,7 @@ def _update_rs_from_primary(
                 and server.address != server_description.address):
 
             # Reset old primary's type to Unknown.
-            sds[server.address] = ServerDescription(server.address)
+            sds[server.address] = server.to_unknown()
 
             # There can be only one prior primary.
             break
